<%inherit file="/html/_base.html" />
<table class="result-overview">
  <tbody>
    <tr>
      <th>Package name</th>
      <th>Module name</th>
      <th>Statement coverage</th>
      <th>Decision coverage</th>
      <th>Condition coverage</th>
    </tr>
  </tbody>
  <tbody>
% for package, package_summary in sorted(summary.packages.items()):
    <tr id="package-row-${ package }" class="package-row">
      <td>${ package }</td>
      <td> - </td>
      <td class="percent" >${ self.percent(package_summary.statement_rate()) }</td>
      <td class="percent" >${ self.percent(package_summary.decision_rate()) }</td>
      <td class="percent" >${ self.percent(package_summary.condition_rate()) }</td>
    </tr>
%   for module, module_summary in sorted(package_summary.modules.items()):
    <tr class="module-row">
      <td>${ package }</td>
      <td><a href="${ module }.html">${ module }</a></td>
      <td class="percent" >${ self.percent(module_summary.statement_rate()) }</td>
      <td class="percent" >${ self.percent(module_summary.decision_rate()) }</td>
      <td class="percent" >${ self.percent(module_summary.condition_rate()) }</td>
    </tr>
%   endfor
% endfor
  </tbody>
</table>
