div.source table {
    border-collapse: collapse;
}

div.source tr {
    padding: 0px;
    margin: 0px;
}

div.source tr.hit{
    background: lightgreen;
}

div.source tr.miss{
    background: red;
}

div.source td {
    padding: 0px;
    margin: 0px;
}

div.source td.lineno {
    background: lightgray;
}

div.source td.missed-conditions {
    background: red;
}

div.source td.hoverable {
    background: red;
}

div.source pre {
    padding: 0px;
    margin: 0px;
}

table.result-overview td.percent {
    text-align: right;
}

