<!doctype html>
<html>
  <head>
    <title>instrumental coverage report</title>
    <script type="text/javascript" src="jquery-1.7.1.min.js"></script>
    <script type="text/javascript" src="jquery-ui-1.8.17.custom.min.js"></script>
    <link rel="stylesheet" type="text/css" href="styles.css" />
    <link rel="stylesheet" type="text/css" href="ui-lightness/jquery-ui-1.8.17.custom.css" />
  </head>
  <body>
    ${ next.body() }
  </body>
</html>

<%def name="percent(value)">${ "%.2f" % (value * 100) } %</%def>
